"""
Utilities for extracting tweet related entities from V2 Twitter API JSON as
collected via twarc.

The data processing for this module is built around a set of
[Glom](https://glom.readthedocs.io/en/latest/) specifications for the individual
little pieces of the API data, composed in PAGE_SPEC for processing an
arbitrary page of API data.

Note that the functionality here works with one page of API data at a time -
when processing lots of data you will need to consider how you will handle
problems like users and tweets duplicated over time/API responses. The DB
module and schema are designed to represent this naturally for you in
an SQLite database.

"""

from collections import defaultdict
import json

from glom import glom, Coalesce, Merge, T

# TODO: user profile related entities.
# entities.description.cashtags
# entities.description.hashtags
# entities.description.mentions
# entities.description.urls
# entities.url.urls

USER_SPEC = {
    "id": "id",
    "created_at": "created_at",
    "username": "username",
    "name": "name",
    "description": "description",
    "location": Coalesce("location", default=None),
    "pinned_tweet_id": Coalesce("pinned_tweet_id", default=None),
    "profile_image_url": "profile_image_url",
    "protected": "protected",
    "followers_count": "public_metrics.followers_count",
    "following_count": "public_metrics.following_count",
    "listed_count": "public_metrics.listed_count",
    "tweet_count": "public_metrics.tweet_count",
    "url": "url",
    "verified": "verified",
    "withheld_country_codes": Coalesce(
        ("withheld.country_codes", json.dumps), default=None
    ),
}

# This unpacks the list of dicts to a single dict,
# then generates nulls for not present values.
REFERENCED_TWEET_SPEC = (
    [{T["type"]: T["id"]}],
    Merge(),
    {
        "replied": Coalesce("replied", default=None),
        "retweeted": Coalesce("retweeted", default=None),
        "quoted": Coalesce("quoted", default=None),
    },
)

# This is for the tweet entities - note there are some differences
# for the entities in a profile object
HASHTAGS = CASHTAGS = ["tag"]
TWEET_MENTIONS = [{"username": "username", "user_id": "id"}]
CONTEXT_ANNOTATIONS = [
    {
        "domain_id": "domain.id",
        "domain_name": "domain.name",
        "domain_description": Coalesce("domain.description", default=None),
        "entity_id": "entity.id",
        "entity_name": "entity.name",
        "entity_description": Coalesce("entity.description", default=None),
    }
]

# Note that URLs are heterogenous - I think part of this is driven by whether
# the page at the URL has an appropriate social media sharing card.
TWEET_URLS = [
    {
        "description": Coalesce("description", default=None),
        "display_url": Coalesce("display_url", default=None),
        "expanded_url": Coalesce("expanded_url", default=None),
        "images": Coalesce(("images", json.dumps), default=None),
        "media_key": Coalesce("media_key", default=None),
        "status": Coalesce("status", default=None),
        "title": Coalesce("title", default=None),
        "unwound_url": Coalesce("unwound_url", default=None),
        "url": Coalesce("url", default=None),
    }
]


TWEET_SPEC = (
    # Because of the structure of the referenced_tweets key,
    # process this into two dicts, that will then be merge later
    # into a single object.
    {
        "all": {
            "id": "id",
            "conversation_id": "conversation_id",
            "author_id": "author_id",
            "created_at": "created_at",
            "text": "text",
            "lang": "lang",
            "source": "source",
            "like_count": "public_metrics.like_count",
            "quote_count": "public_metrics.quote_count",
            "reply_count": "public_metrics.reply_count",
            "retweet_count": "public_metrics.retweet_count",
            "reply_settings": "reply_settings",
            "possibly_sensitive": "possibly_sensitive",
            "withheld_copyright": Coalesce("withheld.copyright", default=None),
            "withheld_country_codes": Coalesce(
                ("withheld.country_codes", json.dumps), default=None
            ),
            # Grab the first poll id only - currently a tweet can only have one
            # poll, but the data structure implies 1:many relationship
            "poll_id": Coalesce(("attachments.poll_ids", T[0]), default=None),
            "place_id": Coalesce("geo.place_id", default=None),
            "media_keys": Coalesce("attachments.media_keys", default=[]),
            "hashtags": (Coalesce("entities.hashtags", default=[]), HASHTAGS),
            "cashtags": (Coalesce("entities.cashtags", default=[]), CASHTAGS),
            "mentions": (Coalesce("entities.mentions", default=[]), TWEET_MENTIONS),
            "urls": (Coalesce("entities.urls", default=[]), TWEET_URLS),
            "context_annotations": (
                Coalesce("context_annotations", default=[]),
                CONTEXT_ANNOTATIONS,
            ),
        },
        "referenced": Coalesce(
            ("referenced_tweets", REFERENCED_TWEET_SPEC),
            default={"replied": None, "retweeted": None, "quoted": None},
        ),
    },
    # Unpack the subdictionaries
    T.values(),
    # Then merge them together
    Merge(),
)

POLL_SPEC = {
    "duration_minutes": "duration_minutes",
    "end_datetime": "end_datetime",
    "id": "id",
    "options": "options",
    "voting_status": "voting_status",
}

# TODO: entity annotations
# "entities.annotations": "entities.annotations",


PLACE_SPEC = {
    "country": "country",
    "country_code": "country_code",
    "full_name": "full_name",
    "geo_type": "geo.type",
    "geo_bbox_1": ("geo.bbox", T[0]),
    "geo_bbox_2": ("geo.bbox", T[1]),
    "geo_bbox_3": ("geo.bbox", T[2]),
    "geo_bbox_4": ("geo.bbox", T[3]),
    # TODO: investigate how important the properties are
    "id": "id",
    "name": "name",
    "place_type": "place_type",
}

MEDIA_SPEC = {
    "alt_text": Coalesce("alt_text", default=None),
    "duration_ms": Coalesce("duration_ms", default=None),
    "media_key": "media_key",
    "preview_image_url": Coalesce("preview_image_url", default=None),
    "view_count": Coalesce("view_count", default=None),
    "type": "type",
    "url": Coalesce("url", default=None),
    "width": "width",
    "height": "height",
}

METADATA_SPEC = {
    "twarc_version": "version",
    "retrieved_at": "retrieved_at",
    "twitter_url": "url",
}

# This is a composition of all of the individual specs into a single glom spec
# that should work on the different variations of responses from the Twitter API.
PAGE_SPEC = {
    "metadata": ("__twarc", METADATA_SPEC),
    # Note that only one of these can match, the main payload can't be of mixed type.
    "data": Coalesce(
        # User data type
        ({"users": ("data", [USER_SPEC])}),
        # Array of tweets (search, timelines etc)
        ({"tweets": ("data", [TWEET_SPEC])}),
        # Single tweet - sample/filter - the lambda is necessary to take the
        # output and wrap in an array so that it works like all the other
        # data structures.
        ({"tweets": ("data", TWEET_SPEC, lambda x: [x])}),
        # Note that no default is specified here - if nothing matches an error needs to be raised
        # because otherwise we don't know what the data actually is.
    ),
    "includes": {
        "users": (Coalesce("includes.users", default=[]), [USER_SPEC]),
        "tweets": (Coalesce("includes.tweets", default=[]), [TWEET_SPEC]),
        "polls": (Coalesce("includes.polls", default=[]), [POLL_SPEC]),
        "places": (Coalesce("includes.places", default=[]), [PLACE_SPEC]),
        "media": (Coalesce("includes.media", default=[]), [MEDIA_SPEC]),
    },
}


def process_page(raw_page):

    return glom(json.loads(raw_page), PAGE_SPEC)


def process_pages(raw_pages):

    processed_pages = []

    for raw_page in raw_pages:
        page = json.loads(raw_page)

        processed_pages.append(glom(page, PAGE_SPEC))

    return processed_pages



import csv
import json

import click

from . import db
from . import twittersphere as twt


@click.group()
@click.pass_context
def twittersphere(ctx):
    """
    Twittersphere
    """


@twittersphere.command("prepare")
@click.argument("input_files", type=click.Path(exists=True), nargs=-1)
@click.argument("output_db", type=click.Path())
@click.option("--n_cpus", default=2, type=click.IntRange(min=1))
@click.pass_context
def prepare(ctx, input_files, output_db, n_cpus):
    """
    Processes the Twitter V2 JSON input_files into the output_db.

    """

    def iterate_pages(input_files):
        for input_file in input_files:
            with open(input_file, "rb") as f:
                for line in f:
                    yield line

    db.insert_pages(output_db, iterate_pages(input_files), n_cpus=n_cpus)


@twittersphere.command("filter-users")
@click.argument("rule-file", type=click.Path(exists=True))
@click.argument("prepared-db", type=click.Path(exists=True))
@click.option("--ruleset-name", default=None, type=str)
# @click.option("--qa-profile-count", type=int, default=100)
@click.pass_context
def filter_users(ctx, rule_file, prepared_db, ruleset_name):
    """
    Apply the given ruleset to user profiles in `prepared-db`.

    The latest `retrieved-at` version of each user profile will be used for
    labelling. The `user_id` of each matching profile will be inserted
    into the `user_matching_ruleset` table with the name of the ruleset.
    This can be customised with the `--ruleset-name`, or if not provided,
    the filename of the `rule-file` will be used instead.

    If `ruleset_name` already exists in the table, it will be replaced.

    """

    db_conn = db.ensure_db(prepared_db)
    include_rules, exclude_rules, _ = twt.load_rule_files([rule_file])

    twt.apply_user_rules_to_db(
        db_conn, include_rules, exclude_rules, ruleset_name or rule_file
    )


@twittersphere.command("filter")
@click.argument("input_files", type=click.Path(exists=True), nargs=-1)
@click.argument("output_file", type=click.Path(exists=False))
@click.option("--rules", type=click.Path(exists=True))
@click.option("--qa-profile-count", type=int, default=100)
@click.pass_context
def filter(ctx, input_files, output_file, rules, qa_profile_count):

    # Instantiate and validate the rules
    include_rules, exclude_rules, all_rules = load_rule_files([rules])
    total_include_rules = sum(len(x) for x in include_rules.values())
    total_exclude_rules = sum(len(x) for x in exclude_rules.values())

    print(
        f"Loaded {total_include_rules} distinct inclusion rules "
        f"and {total_exclude_rules} distinct exclusion rules from {rules}"
    )

    tweets_seen = 0
    tweets_matched = 0
    next_status = 10000

    with open(output_file, "w") as out:

        account_tweet_count = dict()

        for file in input_files:
            with open(file, "r") as f:
                for line in f:
                    api_result_page = json.loads(line)
                    transformed_result_page, page_account_tweet_counts = apply_rules(
                        api_result_page, include_rules, exclude_rules
                    )

                    tweets_seen += len(api_result_page["data"])
                    tweets_matched += len(transformed_result_page["data"])

                    # Merge the count of tweets by account. We keep track of the distinct
                    # versions of the profile information for reporting purposes.
                    for (
                        (matched, author_id),
                        (profile_fields, tweet_count),
                    ) in page_account_tweet_counts.items():

                        if profile_fields in account_tweet_count:
                            account_tweet_count[profile_fields][0].add(author_id)
                            account_tweet_count[profile_fields][1] += tweet_count

                        else:
                            account_tweet_count[profile_fields] = [
                                set([author_id]),
                                tweet_count,
                                matched,
                            ]

                    if tweets_seen >= next_status:
                        print(f"Matched {tweets_matched}/{tweets_seen}.")

                        next_status += 10000

                    out.write(json.dumps(transformed_result_page))
                    out.write("\n")

    print(f"Matched {tweets_matched}/{tweets_seen} in total.")

    qa_path = f"{output_file}.qa.csv"
    print(f"Saving QA data to {qa_path}")

    with open(qa_path, "w") as qa:
        writer = csv.writer(qa, quoting=csv.QUOTE_ALL, dialect="excel")

        writer.writerow(
            [
                "name",
                "description",
                "location",
                "unique_profiles",
                "tweet_count",
                "obfuscated_match",
                "human_label",
            ]
        )
        for row in transform_profiles_for_annotation(
            account_tweet_count, qa_profile_count
        ):
            writer.writerow(row)


@twittersphere.command("concordance")
@click.argument("input_file", type=click.Path(exists=True))
@click.pass_context
def concordance(ctx, input_file):
    """
    Measure concordance between the automated labels and the human labels.

    TODO:

    The output file deobfuscates the machine matched column and rules, and
    creates additional columns describing the entries in the concordance
    table.

    """
    with open(input_file, "r") as f:
        reader = csv.reader(f, quoting=csv.QUOTE_ALL, dialect="excel")

        # skip header
        next(reader)

        bit_location = 2**16

        # Accumulators
        concordance = Counter()
        weighted_concordance = Counter()

        for row in reader:
            profile_count = int(row[3])
            tweet_count = int(row[4])
            auto_match = bool(int(row[5], 0) & bit_location)
            human_match = bool(row[6])

            concordance[(human_match, auto_match)] += profile_count
            weighted_concordance[(human_match, auto_match)] += tweet_count

        print(
            "Profile weighted concordance counts. Format is Human/Auto decision: weight."
        )
        for decision, weight in sorted(concordance.items()):
            print(decision, weight)

        print("Tweet weighted concordance counts.")
        for decision, weight in sorted(weighted_concordance.items()):
            print(decision, weight)



"""
The schema creation statements, and the associated index and materialised view
statements to be run after inserting data.

TODO:

- Likes, followers, following.

"""

CURRENT_SCHEMA_VERSION = 12

SCHEMA_STATEMENTS = """
CREATE table if not exists collection_context (
    context_id integer primary key,
    retrieved_at datetime,
    twitter_url text,
    twarc_version text,
    unique (retrieved_at, twitter_url, twarc_version)
);

CREATE table if not exists user_at_time (
/*
Design notes:
- This schema is an attempt to provide affordances for analytical work, while still
  representing as much as possible as we understand of the underlying data.
- All tables have defined primary keys and appropriate foreign key constraints
  that will be enforced by the database
- Selective denormalisation is applied on most of the 'narrow' tables, to
  allow for certain operations to be performed without a join, or to allow
  indexes to support some anticipated common operations. The columns added
  (where appropriate) are: user_id, created_at, to support trend and user
  counts respectively.
- A set of indexes is suggested to support analytical queries
- The directly collected table is intended to capture the case when a tweet is
  directly retrieved as part of an operation, so as to differentiate direct
  vs referential capture. It is modelled as an append only table, that can
  be joined with the main tweet table to construct the differentiation and to
  simplify the insert logic.
Examples of anticipated queries and purposes are shown along with each query,
and some more extended examples below.
*/
    user_id integer,
    context_id integer references collection_context,
    retrieved_at datetime,
    name text,
    profile_image_url text,
    created_at text,
    protected text,
    description text,
    location text,
    pinned_tweet_id integer,
    verified integer, -- boolean
    url text,
    username text,
    followers_count integer,
    following_count integer,
    tweet_count integer,
    listed_count integer,
    withheld_country_codes text,
    -- TODO: withheld related information
    primary key (user_id, retrieved_at)
) without rowid;

create table if not exists directly_collected_user(
    /*
    Records the IDs of users that were 'directly' collected as part of a collection.
    This allows distinguishing between a user that was included by directly from an
    an API call, or was only referenced by another collection.

    Examples:
    -- The number of directly collected users in this collection
    select
        count(*)
    from directly_collected_user;
    */
    user_id integer primary key
);

create table if not exists tweet_at_time (
    /*
    Contains the 'immutable' characteristics of a tweet, such as the text,
    structure of the conversation, the user ID and so on.
    Mutable characteristics of a tweet are captured in tweet metrics, to
    allow natural capture of change over time without complicating the
    main tweet table. Other aspects of tweets such as hashtags and
    mentions are treated as insert only tables of attributes to handle
    change over time.
    Examples:
    -- Count the number of tweets
    select count(*) from tweet;
    -- Select tweets that were 'directly' collected
    select count(*)
    from tweet
    inner join directly_collected_tweet using(tweet_id);
    -- Find the number of tweets and number of unique users
    -- by month that were directly collected
    select
        datetime(created_at, 'start of month') as month,
        count(distinct user_id) as active_users,
        count(*) as tweets
    from tweet
    inner join directly_collected_tweet using(tweet_id);
    -- Find the ten largest threads by number of tweets
    select
        conversation_id,
        count(*) as tweet_count
    from tweet
    group by conversation_id
    order by tweet_count desc limit 10;
    */
    tweet_id integer,
    context_id integer references collection_context,
    user_id integer,
    created_at text,
    retrieved_at datetime,
    conversation_id integer,
    retweeted_tweet_id integer,
    quoted_tweet_id integer,
    replied_to_tweet_id integer,
    text text,
    lang text,
    source text,
    possibly_sensitive integer, -- boolean
    reply_settings text,
    like_count integer,
    quote_count integer,
    reply_count integer,
    retweet_count integer,
    withheld_copyright integer, --boolean
    withheld_country_codes text,
    poll_id integer,
    place_id integer references place,
    primary key (tweet_id, retrieved_at),
    foreign key (user_id, retrieved_at) references user_at_time,
    foreign key (poll_id, retrieved_at) references poll
) without rowid;

create table if not exists directly_collected_tweet(
    /*
    Records the IDs of tweets that were 'directly' collected as part of a
    collection. This allows distinguishing between a tweet that was included
    by directly from an an API call, such as matching a search query, or was
    only referenced by another tweet such as by quoting.

    Examples:
    -- The number of directly collected tweets in this collection
    select
        count(*)
    from directly_collected_tweet;
    */
    tweet_id integer primary key
);

create table if not exists tweet_hashtag(
    tweet_id integer,
    retrieved_at datetime,
    hashtag text,
    foreign key (tweet_id, retrieved_at) references tweet_at_time,
    primary key (tweet_id, retrieved_at, hashtag)
) without rowid;

create table if not exists tweet_cashtag(
    tweet_id integer,
    retrieved_at datetime,
    cashtag text,
    foreign key (tweet_id, retrieved_at) references tweet_at_time,
    primary key (tweet_id, retrieved_at, cashtag)
) without rowid;

create table if not exists tweet_mention(
    tweet_id integer,
    retrieved_at datetime,
    mentioned_user_id text,
    mentioned_username text,
    foreign key (tweet_id, retrieved_at) references tweet_at_time,
    primary key (tweet_id, retrieved_at, mentioned_user_id)
) without rowid;

create table if not exists url(
    url text,
    retrieved_at datetime,
    description text,
    display_url text,
    expanded_url text,
    images text,
    media_key text,
    status text,
    title text,
    unwound_url text,
    primary key (url, retrieved_at)
) without rowid;

create table if not exists tweet_url(
    tweet_id integer,
    retrieved_at datetime,
    url text,
    foreign key (tweet_id, retrieved_at) references tweet_at_time,
    foreign key (url, retrieved_at) references url,
    primary key (tweet_id, retrieved_at, url)
) without rowid;

create table if not exists poll(
    poll_id integer,
    retrieved_at datetime,
    duration_minutes integer,
    end_datetime datetime,
    voting_status text,
    primary key (poll_id, retrieved_at)
) without rowid;

create table if not exists poll_option(
    poll_id integer,
    retrieved_at datetime,
    position integer,
    label text,
    votes integer,
    primary key (poll_id, retrieved_at, position),
    foreign key (poll_id, retrieved_at) references poll
) without rowid;

create table if not exists place(
    place_id text primary key,
    country text,
    country_code text,
    full_name text,
    geo_type text,
    geo_bbox_1 float,
    geo_bbox_2 float,
    geo_bbox_3 float,
    geo_bbox_4 float,
    name text,
    place_type text
) without rowid;

create table if not exists media(
    media_key text,
    retrieved_at datetime,
    alt_text text,
    duration_ms integer,
    preview_image_url text,
    view_count integer,
    type text,
    url text,
    width integer,
    height integer,
    primary key (media_key, retrieved_at)
) without rowid;

create table if not exists tweet_media(
    tweet_id integer,
    retrieved_at datetime,
    media_key text,
    primary key (tweet_id, retrieved_at, media_key)
    foreign key (media_key, retrieved_at) references media
) without rowid;

create table if not exists domain(
    domain_id primary key,
    name text,
    description text
);

create table if not exists entity(
    entity_id primary key,
    name text,
    description text
);

create table if not exists tweet_entity_domain(
    tweet_id integer,
    retrieved_at datetime,
    entity_id integer references entity,
    domain_id integer references domain,
    primary key (tweet_id, retrieved_at, entity_id, domain_id)
) without rowid;

create table if not exists user_matching_ruleset(
    ruleset_name text,
    user_id integer,
    primary key (ruleset_name, user_id)
);

create table if not exists user_ruleset_ngram_count(
    /* The empty string ruleset name will be interpreted as the global count
    of all profiles. */
    ruleset_name text,
    field text,
    first_token text,
    second_token text,
    third_token text,
    profile_count integer default 0,
    primary key (ruleset_name, field, first_token, second_token, third_token)
) without rowid;


create table if not exists metadata (
    key text primary key,
    value
);

create view if not exists user_latest as
    /* Note that this relies on the SQLite behaviour of selecting the row
       corresponding to the max value in the case of the group by. */
    select
        max(retrieved_at) as latest_retrieved_at,
        *
    from user_at_time
    group by user_id;

create view if not exists tweet_latest as
    /* Note that this relies on the SQLite behaviour of selecting the row
       corresponding to the max value in the case of the group by. */
    select
        max(retrieved_at) as latest_retrieved_at,
        *
    from tweet_at_time
    group by tweet_id;



insert or ignore into metadata values('twittersphere_schema_version', {});

""".format(
    CURRENT_SCHEMA_VERSION
)


table_keys = {
    "collection_context": "context_id",
    "user_at_time": "user_id,retrieved_at",
    "directly_collected_user": "user_id",
    "tweet_at_time": "tweet_id,retrieved_at",
    "directly_collected_tweet": "tweet_id",
    "tweet_hashtag": "tweet_id,retrieved_at,hashtag",
    "tweet_cashtag": "tweet_id,retrieved_at,cashtag",
    "tweet_mention": "tweet_id,retrieved_at,mentioned_user_id",
    "url": "url,retrieved_at",
    "tweet_url": "tweet_id,retrieved_at,url",
    "poll": "poll_id,retrieved_at",
    "poll_option": "poll_id,retrieved_at,position",
    "place": "place_id",
    "media": "media_key,retrieved_at",
    "tweet_media": "tweet_id,retrieved_at,media_key",
    "domain": "domain_id",
    "entity": "entity_id",
    "tweet_entity_domain": "tweet_id,retrieved_at,entity_id,domain_id",
    "metadata": "key",
}
